#include "MidRadio.h"
#include <tchar.h>

#ifdef _DEBUG
#define OUTPUT_DEBUG_MSG(x) OutputDebugString(x)
#else
#define OUTPUT_DEBUG_MSG(x)
#endif

CMidRadio::CMidRadio(CRITICAL_SECTION& cs, LPUNKNOWN& pUnkControl, HWND mhWnd) :
  m_cs(cs), mUnkControl(pUnkControl)
{
  ::EnterCriticalSection(&m_cs);

  hwndTop = GetTopLevelWindow(mhWnd);
  if (hwndTop)
  {
    ::EnableWindow(hwndTop, FALSE);
    ::SetCapture(hwndTop);
  }
  hcurPrev = ::SetCursor(::LoadCursor(NULL, IDC_WAIT));

  mr = NULL;
  if (mUnkControl)
  {
    mUnkControl->QueryInterface(IID_IMidRadioCtrl, (void **)&mr);
  }
}

CMidRadio::~CMidRadio()
{
  if (mr)
  {
    mr->Release();
  }

  ::SetCursor(hcurPrev);
  if (hwndTop)
  {
    ::ReleaseCapture();
    ::EnableWindow(hwndTop, TRUE);
  }

  ::LeaveCriticalSection(&m_cs);
}


HWND CMidRadio::GetTopLevelWindow(HWND hWnd)
{
	HWND hWndTop;

	do
	{
		hWndTop = hWnd;
		hWnd = ::GetParent(hWnd);
	}
	while (hWnd != NULL);

	return hWndTop;
}

HRESULT CMidRadio::raw_Play()
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::raw_Play\n"));

  if (mr) return mr->raw_Play();
  return E_NOINTERFACE;
}

HRESULT CMidRadio::raw_Stop()
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::raw_Stop\n"));

  if (mr) return mr->raw_Stop();
  return E_NOINTERFACE;
}

HRESULT CMidRadio::raw_Pause()
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::raw_Pause\n"));

  if (mr) return mr->raw_Pause();
  return E_NOINTERFACE;
}

HRESULT CMidRadio::raw_Next()
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::raw_Next\n"));

  if (mr) return mr->raw_Next();
  return E_NOINTERFACE;
}

HRESULT CMidRadio::raw_Prev()
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::raw_Prev\n"));

  if (mr) return mr->raw_Prev();
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Src(BSTR * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Src\n"));

  if (mr) return mr->get_Src(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_Src(BSTR pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_Src\n"));

  if (mr) return mr->put_Src(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_AutoStart(VARIANT_BOOL * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_AutoStart\n"));

  if (mr) return mr->get_AutoStart(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_AutoStart(VARIANT_BOOL pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_AutoStart\n"));

  if (mr) return mr->put_AutoStart(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Repeat(VARIANT * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Repeat\n"));

  if (mr) return mr->get_Repeat(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_Repeat(VARIANT & pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_Repeat\n"));

  if (mr) return mr->put_Repeat(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Text(BSTR * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Text\n"));

  if (mr) return mr->get_Text(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_Text(BSTR pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_Text\n"));

  if (mr) return mr->put_Text(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Volume(long * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Volume\n"));

  if (mr) return mr->get_Volume(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_Volume(long pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_Volume\n"));

  if (mr) return mr->put_Volume(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Tempo(long * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Tempo\n"));

  if (mr) return mr->get_Tempo(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_Tempo(long pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_Tempo\n"));

  if (mr) return mr->put_Tempo(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_TimeFormat(long * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_TimeFormat\n"));

  if (mr) return mr->get_TimeFormat(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_TimeFormat(long pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_TimeFormat\n"));

  if (mr) return mr->put_TimeFormat(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_CurrentPosition(BSTR * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_CurrentPosition\n"));

  if (mr) return mr->get_CurrentPosition(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_CurrentPosition(BSTR pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_CurrentPosition\n"));

  if (mr) return mr->put_CurrentPosition(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_StopPosition(BSTR * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_StopPosition\n"));

  if (mr) return mr->get_StopPosition(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_StopPosition(BSTR pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_StopPosition\n"));

  if (mr) return mr->put_StopPosition(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Duration(BSTR * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Duration\n"));

  if (mr) return mr->get_Duration(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_Duration(BSTR pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_Duration\n"));

  if (mr) return mr->put_Duration(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Transpose(long * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Transpose\n"));

  if (mr) return mr->get_Transpose(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::put_Transpose(long pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::put_Transpose\n"));

  if (mr) return mr->put_Transpose(pVal);
  return E_NOINTERFACE;
}

HRESULT CMidRadio::get_Version(BSTR * pVal)
{
  OUTPUT_DEBUG_MSG(_T("CMidRadio::get_Version\n"));

  if (mr) return mr->get_Version(pVal);
  return E_NOINTERFACE;
}
