
#include "plugin.h"
#include "MidRadio.tlh"

CMidRadioEventListener::CMidRadioEventListener(nsPluginInstance* aPlugin)
{
  m_refCount = 0;
  mPlugin = aPlugin;
}

CMidRadioEventListener::~CMidRadioEventListener()
{
}

// IUnknown methods
STDMETHODIMP CMidRadioEventListener::QueryInterface(REFIID riid,
                                                    void ** ppvObj)
{
  if (IID_IUnknown == riid)
  {
    *ppvObj = static_cast<IUnknown*>(this);
  }
  else if (IID_IDispatch == riid)
  {
    *ppvObj = static_cast<IDispatch*>(this);
  }
  else if (DIID__IMidRadioCtrlEvents == riid)
  {
    *ppvObj = static_cast<IDispatch*>(this);
  }
  else
  {
    *ppvObj = NULL;
    return E_NOINTERFACE;
  }

  static_cast<IUnknown*>(*ppvObj)->AddRef();
  return S_OK;
}

STDMETHODIMP_(ULONG) CMidRadioEventListener::AddRef()
{
  return ++m_refCount;
}

STDMETHODIMP_(ULONG) CMidRadioEventListener::Release()
{
  if (--m_refCount == 0)
  {
    delete this;
    return 0;
  }
  return m_refCount;
}

// IDispatch methods
STDMETHODIMP CMidRadioEventListener::GetTypeInfoCount(UINT *iTInfo)
{
  *iTInfo = 0;
  return S_OK;
}

STDMETHODIMP CMidRadioEventListener::GetTypeInfo(UINT iTInfo,
                                                 LCID lcid,
                                                 ITypeInfo **ppTInfo)
{
  return E_NOTIMPL;
}

STDMETHODIMP CMidRadioEventListener::GetIDsOfNames(REFIID riid,
                                                   OLECHAR **rgszNames,
                                                   UINT cNames,
                                                   LCID lcid,
                                                   DISPID *rgDispId)
{
  return E_NOTIMPL;
}

STDMETHODIMP CMidRadioEventListener::Invoke(DISPID dispIdMember,
                                            REFIID riid,
                                            LCID lcid,
                                            WORD wFlags,
                                            DISPPARAMS* pDispParams,
                                            VARIANT* pVarResult,
                                            EXCEPINFO* pExcepInfo,
                                            UINT* puArgErr)
{
  if (IID_NULL != riid)
    return E_INVALIDARG;

  HRESULT hr = S_OK;

  if (mPlugin)
  {
    switch (dispIdMember)
    {
    case 1:
      mPlugin->OnEnd();
      break;

    case 2:
      mPlugin->OnEndClip();
      break;

    case 3:
      mPlugin->OnStart();
      break;

    case 4:
      mPlugin->OnStartClip();
      break;

    case 5:
      mPlugin->OnPause();
      break;

    case 6:
      mPlugin->OnNext();
      break;

    case 7:
      mPlugin->OnPrev();
      break;

    case 8:
      mPlugin->OnRestart();
      break;

    case 9:
      mPlugin->OnClickPlay();
      break;

    case 10:
      mPlugin->OnClickStop();
      break;

    case 11:
      mPlugin->OnClickPause();
      break;

    case 12:
      mPlugin->OnClickNext();
      break;

    case 13:
      mPlugin->OnClickPrev();
      break;
    }
  }
  return hr;
}
