/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is 
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
#include "nsStringAPI.h"
#include "nsIComponentManager.h"
#include "nsILocalFile.h"
#include "nsXPCOMCID.h"
#include "nsIURL.h"
*/
#include <nsNetUtil.h>

#include <windows.h>
#include <windowsx.h>
#include <wininet.h>
#include <shlwapi.h>
#include <imagehlp.h>
#include <intrin.h>
#include <tchar.h>
#include <strsafe.h>
#include <crtdbg.h>

#define ATL_AXCREATECONTROLEX m_pfnAtlAxCreateControlEx

/*
#include <atlcoll.h>
#include <atlhost.h>

class CNPMRHLPModule : public CAtlDllModuleT< CNPMRHLPModule >
{
public :
	// DECLARE_LIBID(LIBID_NPMRHLPLib)
};
CNPMRHLPModule _AtlModule;
#undef  ATL_AXCREATECONTROLEX
#define ATL_AXCREATECONTROLEX AtlAxCreateControlEx
*/

#include "plugin.h"

#ifdef NPMRHLP_SCRIPTABLE
#include "ScriptablePluginObject.h"
#include "nsIVariant.h"
#include "npruntime.h"
#endif

#include "MidRadio.h"
#pragma comment(linker, "/merge:.rdata=.text")


#define SRC_NULL_TERMINATED (-1)

char* NPN_StrDuplicate(const char* s);
BSTR AllocSysString(const char *pchData, int cbMultiByte, UINT CodePage);
BSTR AllocSysString(const wchar_t *pchData, int cchWideChar);
char* NPN_AllocString(LPCWSTR pchData, int cchWideChar, UINT CodePage);
LPWSTR AllocWString(LPCSTR pchData, int cbMultiByte, UINT CodePage);
BSTR AllocSysStringFromUTF8(LPCSTR lpszUTF8, int cbMultiByte);
bool UTF8ToLong(const char *lpszUtf8, int cbMultiByte, LONG *plOut);
bool UTF8ToDouble(const char *lpszUTF8, int cbMultiByte, double *pdOut);
void DrawPluginErrMsg(HWND hWnd, HDC hdc, LPPAINTSTRUCT lpps, const TCHAR *lpszErr);

// global variables
static const TCHAR szMidRadioContainerTT[] = _T("MidRadioContainerTT");
static const TCHAR PLUGIN_INSTANCE_PTR[] = _T("PluginInstancePtr");


class CInitializeGlobalData
{
public:
  CInitializeGlobalData()
  {
    m_bComInitialized = SUCCEEDED(::CoInitialize(NULL));
    ::InitializeCriticalSection(&m_CritSec);

    m_pUnkDummyControl = NULL;
    m_pfnHeapAlloc = NULL;
    m_hModAtl = NULL;
    g_hModMidRadio = NULL;
    m_bReplacedMidRadioHeapAlloc = false;
    m_pfnAtlAxCreateControlEx = NULL;

    // register container window class
    WNDCLASS wc;

    wc.style = CS_DBLCLKS;
    wc.lpfnWndProc = ::DefWindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = ::GetModuleHandle(NULL);
    wc.hIcon = NULL;
    wc.hCursor = ::LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szMidRadioContainerTT;

    if (::RegisterClass(&wc))
    {
      // load ATL DLL
      const TCHAR* szKnownAtlDlls[] = {
  /*
        _T("atl110.dll"),
        _T("atl100.dll"),
        _T("atl90.dll"),
        _T("atl80.dll"),
        _T("atl71.dll"),
        _T("atl70.dll"),
  */
        _T("atl.dll")
      };

      for (int i = 0; i < sizeof(szKnownAtlDlls) / sizeof(char*); i++)
      {
        m_hModAtl = ::LoadLibrary(szKnownAtlDlls[i]);
        if (m_hModAtl)
        {
          m_pfnAtlAxCreateControlEx =
            (PFN_AtlAxCreateControlEx)::GetProcAddress(m_hModAtl,
                                                       "AtlAxCreateControlEx");
          if (m_pfnAtlAxCreateControlEx)
          {
            break;
          }
        }
      }

      if (m_pfnAtlAxCreateControlEx)
      {
        // hook HeapAlloc of MidRadio.ocx
        m_pfnHeapAlloc = ::GetProcAddress(::GetModuleHandle(_T("kernel32.dll")),
                                          "HeapAlloc");
        if (m_pfnHeapAlloc)
        {
          // create dummy control for hooking HeapAlloc function of MidRadio.ocx
          if (SUCCEEDED(::CoCreateInstance(CLSID_MidRadioCtrl, NULL,
            CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&m_pUnkDummyControl))))
          {
            g_hModMidRadio = ::GetModuleHandle(_T("MidRadio.ocx"));
            if (g_hModMidRadio)
            {
              PIMAGE_NT_HEADERS pNTHeader = GetModuleNTHeaders(g_hModMidRadio);

              if (pNTHeader &&
                  pNTHeader->OptionalHeader.MajorLinkerVersion < 8)
              {
                ReplaceModuleIAT(g_hModMidRadio, m_pfnHeapAlloc, (FARPROC)HeapAlloc_tt);
                m_bReplacedMidRadioHeapAlloc = true;
              }
            }
          }
        }
      }
    }
  }

  virtual ~CInitializeGlobalData()
  {
    if (m_pfnHeapAlloc)
    {
      if (m_pUnkDummyControl)
      {
        if (g_hModMidRadio)
        {
          if (m_bReplacedMidRadioHeapAlloc)
          {
            ReplaceModuleIAT(g_hModMidRadio, (FARPROC)HeapAlloc_tt, m_pfnHeapAlloc);
          }
        }
        m_pUnkDummyControl->Release();
      }
    }

    if (m_hModAtl)
    {
      ::FreeLibrary(m_hModAtl);
    }

    ::UnregisterClass(szMidRadioContainerTT, ::GetModuleHandle(NULL));
    ::DeleteCriticalSection(&m_CritSec);
    if (m_bComInitialized)
    {
      ::CoUninitialize();
    }
  }

  HRESULT CreateMidRadioControl(HWND mhWndContainer,
                                IUnknown** ppUnkControl,
                                IUnknown* punkSink)
  {
    HRESULT hr = E_FAIL;

    if (m_pfnAtlAxCreateControlEx)
    {
      LPOLESTR lpszCLSID = NULL;

      if (SUCCEEDED(::StringFromCLSID(CLSID_MidRadioCtrl, &lpszCLSID)))
      {
        hr = ATL_AXCREATECONTROLEX(lpszCLSID,
          mhWndContainer, NULL, NULL, ppUnkControl,
          DIID__IMidRadioCtrlEvents, punkSink);

        ::CoTaskMemFree(lpszCLSID);
      }
    }
    return hr;
  }

  CRITICAL_SECTION & GetCritSec() { return m_CritSec; }
  HMODULE GetMidRadioHandle() { return g_hModMidRadio; }
  bool AtlInitialized() { return NULL != m_pfnAtlAxCreateControlEx; }

private:
  bool m_bComInitialized;
  CRITICAL_SECTION m_CritSec;
  IUnknown *m_pUnkDummyControl;
  FARPROC m_pfnHeapAlloc;
  HMODULE m_hModAtl;
  bool m_bReplacedMidRadioHeapAlloc;
  static HMODULE g_hModMidRadio;
typedef HRESULT (WINAPI * PFN_AtlAxCreateControlEx)(LPCOLESTR, HWND, IStream*, IUnknown**, IUnknown**, REFIID, IUnknown*);
  PFN_AtlAxCreateControlEx m_pfnAtlAxCreateControlEx;

  void ReplaceModuleIAT(HMODULE hModule, FARPROC from, FARPROC to)
  {
    ULONG size;
    PIMAGE_IMPORT_DESCRIPTOR piid = (PIMAGE_IMPORT_DESCRIPTOR)
      ::ImageDirectoryEntryToData((PVOID)g_hModMidRadio, TRUE,
      IMAGE_DIRECTORY_ENTRY_IMPORT, &size);

    if (piid)
    {
      ULONG_PTR ImageBase = (ULONG_PTR)hModule;

      for (; 0 != piid->OriginalFirstThunk; piid++)
      {
        char *pszName = (char *)(ImageBase + piid->Name);
        PIMAGE_THUNK_DATA pThunk;

        for (pThunk = (PIMAGE_THUNK_DATA)(ImageBase + piid->FirstThunk);
             pThunk->u1.Function;
             pThunk++)
        {
          if ((FARPROC)pThunk->u1.Function == from)
          {
            DWORD dwOldProtect;
            if (::VirtualProtect(pThunk, sizeof(FARPROC), PAGE_EXECUTE_READWRITE, &dwOldProtect))
            {
              *(FARPROC*)pThunk = to;
              ::FlushInstructionCache(::GetCurrentProcess(), pThunk, sizeof(FARPROC));
              ::VirtualProtect(pThunk, sizeof(FARPROC), dwOldProtect, &dwOldProtect);
            }
          }
        }
      }
    }
  }

  static PIMAGE_NT_HEADERS GetModuleNTHeaders(HMODULE hMod)
  {
    PIMAGE_NT_HEADERS pinh = NULL;

    if (hMod)
    {
      PIMAGE_DOS_HEADER dosHeader = (PIMAGE_DOS_HEADER)hMod;

      if (!::IsBadReadPtr(dosHeader, sizeof(dosHeader->e_magic)) &&
	        IMAGE_DOS_SIGNATURE == dosHeader->e_magic)
      {
#define MakePtr(cast, ptr, addValue) (cast)((DWORD_PTR)(ptr) + (DWORD_PTR)(addValue))
        PIMAGE_NT_HEADERS pNTHeader = MakePtr(PIMAGE_NT_HEADERS,
                                              dosHeader, dosHeader->e_lfanew);

        if (!::IsBadReadPtr(pNTHeader, sizeof(pNTHeader->Signature)) &&
	          IMAGE_NT_SIGNATURE == pNTHeader->Signature &&
	          IMAGE_NT_OPTIONAL_HDR_MAGIC == pNTHeader->OptionalHeader.Magic)
        {
          pinh = pNTHeader;
        }
      }
    }
    return pinh;
  }

  static bool IsAddressInMidRadioOcx(void *lpAddress)
  {
    bool bResult = false;

    if (g_hModMidRadio)
    {
      PIMAGE_NT_HEADERS pNTHeader = GetModuleNTHeaders(g_hModMidRadio);

      if (pNTHeader)
      {
        if ((ULONG_PTR)lpAddress >= (ULONG_PTR)g_hModMidRadio &&
            (ULONG_PTR)lpAddress <  (ULONG_PTR)g_hModMidRadio +
                                               pNTHeader->OptionalHeader.SizeOfImage)
        {
          bResult = true;
        }
      }
    }
    return bResult;
  }

  // hooked HeapAlloc
  static LPVOID WINAPI HeapAlloc_tt(HANDLE hHeap, DWORD dwFlags, SIZE_T dwBytes)
  {
    LPVOID lResult = ::HeapAlloc(hHeap, dwFlags, dwBytes);

    if (NULL != lResult &&
        IsAddressInMidRadioOcx(_ReturnAddress()))
    {
      DWORD dwHeapSize = ::HeapSize(hHeap, 0, lResult);

      if (dwHeapSize > 0)
      {
        DWORD dwOldProtect;
        ::VirtualProtect(lResult, dwHeapSize, PAGE_EXECUTE_READWRITE, &dwOldProtect);
      }
    }
    return lResult;
  }

}; 
HMODULE CInitializeGlobalData::g_hModMidRadio = NULL;  
static CInitializeGlobalData igd;


//////////////////////////////////////
//
// general initialization and shutdown
//
NPError NS_PluginInitialize()
{
  return NPERR_NO_ERROR;
}

void NS_PluginShutdown()
{
}

/////////////////////////////////////////////////////////////
//
// construction and destruction of our plugin instance object
//
nsPluginInstanceBase * NS_NewPluginInstance(nsPluginCreateData * aCreateDataStruct)
{
  if(!aCreateDataStruct)
    return NULL;

  nsPluginInstance * plugin = new nsPluginInstance(aCreateDataStruct->instance);
  if (plugin)
  {
    plugin->SetPluginCreateData(aCreateDataStruct);
  }
  return plugin;
}

void NS_DestroyPluginInstance(nsPluginInstanceBase * aPlugin)
{
  if(aPlugin)
    delete (nsPluginInstance *)aPlugin;
}

////////////////////////////////////////
//
// nsPluginInstance class implementation
//

#define DECLARE_NPN_STRINGID(name) \
  m_id_##name = NPN_GetStringIdentifier(#name)

nsPluginInstance::nsPluginInstance(NPP aInstance) : nsPluginInstanceBase()
{
  m_bInitialized = FALSE;
  m_pInstance = aInstance;

  // get window object
  sWindowObj = NULL;
  NPN_GetValue(aInstance, NPNVWindowNPObject, &sWindowObj);

  m_hWnd = NULL;
  m_orgPluginWndProc = NULL;

  m_hWndContainer = NULL;
  m_pUnkControl = NULL;
  m_bComInit = false;

  m_pMidRadioEventListener = NULL;
#ifdef NPMRHLP_SCRIPTABLE
  m_pScriptableObject = NULL;
#endif

  m_bEmbedded = false;

  // embedded data
  ::ZeroMemory(&m_Embed, sizeof(m_Embed));

#ifdef NPMRHLP_SCRIPTABLE
  // Methods
  DECLARE_NPN_STRINGID(Play);
  DECLARE_NPN_STRINGID(Stop);
  DECLARE_NPN_STRINGID(Pause);
  DECLARE_NPN_STRINGID(Next);
  DECLARE_NPN_STRINGID(Prev);

  // Properties
  DECLARE_NPN_STRINGID(Src);
  DECLARE_NPN_STRINGID(AutoStart);
  DECLARE_NPN_STRINGID(Repeat);
  DECLARE_NPN_STRINGID(Text);
  DECLARE_NPN_STRINGID(Volume);
  DECLARE_NPN_STRINGID(Tempo);
  DECLARE_NPN_STRINGID(TimeFormat);
  DECLARE_NPN_STRINGID(CurrentPosition);
  DECLARE_NPN_STRINGID(StopPosition);
  DECLARE_NPN_STRINGID(Duration);
  DECLARE_NPN_STRINGID(Transpose);
  DECLARE_NPN_STRINGID(Version);
#endif
}

#define RELEASE_EMBED_DATA(name) \
  if (m_Embed.m_p##name) NPN_MemFree(m_Embed.m_p##name);

nsPluginInstance::~nsPluginInstance()
{
  if (sWindowObj)
  {
    NPN_ReleaseObject(sWindowObj);
    sWindowObj = NULL;
  }

#ifdef NPMRHLP_SCRIPTABLE
  if (m_pScriptableObject)
  {
    NPN_ReleaseObject(m_pScriptableObject);
    m_pScriptableObject = NULL;
  }
#endif

  RELEASE_EMBED_DATA(AutoStart);
  RELEASE_EMBED_DATA(Repeat);
  RELEASE_EMBED_DATA(Text);
  RELEASE_EMBED_DATA(Volume);
  RELEASE_EMBED_DATA(Tempo);
  RELEASE_EMBED_DATA(TimeFormat);
  RELEASE_EMBED_DATA(CurrentPosition);
  RELEASE_EMBED_DATA(StopPosition);
  RELEASE_EMBED_DATA(Duration);
  RELEASE_EMBED_DATA(Transpose);
}

#define SET_EMBED_DATA(name)                              \
  if (_stricmp(lppcd->argn[i], #name) == 0) {             \
    if (m_Embed.m_p##name)                                \
      NPN_MemFree(m_Embed.m_p##name);                     \
    m_Embed.m_p##name = NPN_StrDuplicate(lppcd->argv[i]); \
  }

void nsPluginInstance::SetPluginCreateData(nsPluginCreateData * lppcd)
{
  m_bEmbedded = (NP_EMBED == lppcd->mode);
  if (m_bEmbedded)
  {
    for (int i = 0; i < lppcd->argc; i++)
    {
      if (lppcd->argv[i] && ::lstrlenA(lppcd->argv[i]) <= 1024)
      {
        SET_EMBED_DATA(AutoStart)
        else SET_EMBED_DATA(Repeat)
        else SET_EMBED_DATA(Text)
        else SET_EMBED_DATA(Volume)
        else SET_EMBED_DATA(Tempo)
        else SET_EMBED_DATA(TimeFormat)
        else SET_EMBED_DATA(CurrentPosition)
        else SET_EMBED_DATA(StopPosition)
        else SET_EMBED_DATA(Duration)
        else SET_EMBED_DATA(Transpose)
      }
    }
  }
}

NPBool nsPluginInstance::init(NPWindow* aWindow)
{
  if(aWindow == NULL)
    return FALSE;

  m_hWnd = (HWND)aWindow->window;
  if(m_hWnd == NULL || !::IsWindow(m_hWnd))
    return FALSE;

  // create MidRadioCtrl
  RECT rc = {0};
  ::GetClientRect(m_hWnd, &rc);
  m_hWndContainer = ::CreateWindow(
    szMidRadioContainerTT,
    _T(""),
    WS_CHILD | WS_VISIBLE,
    0,
    0,
    rc.right - rc.left,
    rc.bottom - rc.top,
    m_hWnd,
    NULL,
    ::GetModuleHandle(NULL),
    NULL);

  if (m_hWndContainer)
  {
    m_bComInit = SUCCEEDED(::CoInitialize(NULL));

    m_pMidRadioEventListener = new CMidRadioEventListener(this);
    m_pMidRadioEventListener->AddRef();

    // create MidRadio ActiveX control and set host.
    if (SUCCEEDED(igd.CreateMidRadioControl(m_hWndContainer,
                  &m_pUnkControl, m_pMidRadioEventListener)))
    {
      InitializeMidRadioAttributes();
    }
    else
    {
      m_pMidRadioEventListener->Release();
      m_pMidRadioEventListener = NULL;
    }
  }

  if (!igd.AtlInitialized() ||
    NULL == m_pUnkControl)
  {
    if (m_hWndContainer)
    {
      ::DestroyWindow(m_hWndContainer);
      m_hWndContainer = NULL;
    }

    // subclass window so we can intercept window messages and
    // do our drawing to it
    m_orgPluginWndProc = SubclassWindow(m_hWnd, (WNDPROC)PluginWinProc);
    if (m_orgPluginWndProc)
    {
      // associate window with our nsPluginInstance object so we can access 
      // it in the window procedure
      ::SetProp(m_hWnd, PLUGIN_INSTANCE_PTR, (HANDLE)this);
    }
  }

  m_bInitialized = TRUE;
  return TRUE;
}

void nsPluginInstance::shut()
{
  if (m_pUnkControl)
  {
    CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl);
    mMidRadio.raw_Stop();

    m_pUnkControl->Release();
    m_pUnkControl = NULL;
  }

  if (m_orgPluginWndProc)
  {
    // subclass it back
    SubclassWindow(m_hWnd, m_orgPluginWndProc);
    m_orgPluginWndProc = NULL;
    ::RemoveProp(m_hWnd, PLUGIN_INSTANCE_PTR);
  }

  if (m_pMidRadioEventListener)
  {
    m_pMidRadioEventListener->Release();
    m_pMidRadioEventListener = NULL;
  }

  if (m_hWndContainer)
  {
    HCURSOR hCursor = ::SetCursor(::LoadCursor(NULL, MAKEINTRESOURCE(IDC_WAIT)));

    ::Sleep(1500);

    if (hCursor)
    {
      ::SetCursor(hCursor);
    }

    ::DestroyWindow(m_hWndContainer);
    m_hWndContainer = NULL;
  }

  if (m_bComInit)
  {
    ::CoUninitialize();
    m_bComInit = false;
  }

  m_hWnd = NULL;
  m_bInitialized = FALSE;
}

NPBool nsPluginInstance::isInitialized()
{
  return m_bInitialized;
}

NPError nsPluginInstance::SetWindow(NPWindow* pNPWindow)
{
  if (m_bInitialized && ::IsWindow(m_hWndContainer) && (pNPWindow->window != NULL))
  {
    // windows resized
    ::SetWindowPos(m_hWndContainer, NULL, 0, 0, pNPWindow->width, pNPWindow->height,
      SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOREDRAW);
  }
  return NPERR_NO_ERROR;
}

NPError nsPluginInstance::NewStream(NPMIMEType type, NPStream* stream,
                                    NPBool seekable, uint16* stype)
{
  if (stype)
  {
    *stype = NP_NORMAL;
  }
  return NPERR_NO_ERROR;
}

int32_t nsPluginInstance::WriteReady(NPStream *stream)
{
  if (m_pInstance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;

  if (stream)
    NPN_DestroyStream(m_pInstance, stream, NPRES_DONE);

  if (m_bInitialized && m_pUnkControl && stream && stream->url)
  {
    int nUrlLen = strlen(stream->url);
    char *lpszDecodedURL = new char[nUrlLen + 1];
    char szTmp[3];
    int iTmp;
    int i, j;

    for (i = 0, j = 0; i < nUrlLen; i++, j++)
    {
      if (stream->url[i] == '%' &&
          i + 2 < nUrlLen)
      {
        szTmp[0] = stream->url[i+1];
        szTmp[1] = stream->url[i+2];
        szTmp[2] = '\0';
        sscanf_s(szTmp, "%02x", &iTmp);
        lpszDecodedURL[j] = iTmp;
        i += 2;
      }
      else
      {
        lpszDecodedURL[j] = stream->url[i];
      }
    }
    lpszDecodedURL[j] = '\0';

    BSTR bs = AllocSysString(lpszDecodedURL, SRC_NULL_TERMINATED, CP_UTF8);
    if (bs)
    {
      CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);

      if (UrlIsFileUrlW(bs))
      {
        WCHAR wszLocalPath[MAX_PATH];
        DWORD cchPath = ::lstrlenW(bs);

        if (cchPath <= INTERNET_MAX_URL_LENGTH &&
            SUCCEEDED(::PathCreateFromUrlW(bs, wszLocalPath, &cchPath, NULL)))
        {
          BSTR bsLocal = AllocSysString(wszLocalPath, SRC_NULL_TERMINATED);
          if (bsLocal)
          {
            mMidRadio.put_Src(bsLocal);
            ::SysFreeString(bsLocal);
          }
        }
      }
      else
      {
        mMidRadio.put_Src(bs);
      }
      ::SysFreeString(bs);
    }
    delete [] lpszDecodedURL;
  }

  return -1L;   // dont accept any bytes in NPP_Write()
}

bool GetBoolFromUTF8(const char *lpszUTF8, bool *_retval)
{
  bool bResult = false;

  if (lpszUTF8 && _retval)
  {
    double d;

    if (*lpszUTF8 == '\0' ||
        _stricmp(lpszUTF8, "false") == 0 ||
        (UTF8ToDouble(lpszUTF8, SRC_NULL_TERMINATED, &d) && d == 0.0))
    {
      *_retval = false;
      bResult = true;
    }
    else
    {
      *_retval = true;
      bResult = true;
    }
  }
  return bResult;
}

#define INIT_MRATTR_WITH_BSTR(name)                                       \
  if (m_Embed.m_p##name)                                                  \
  {                                                                       \
    bs = AllocSysStringFromUTF8(m_Embed.m_p##name, SRC_NULL_TERMINATED);  \
    if (bs)                                                               \
    {                                                                     \
      mMidRadio.put_##name(bs);                                           \
      ::SysFreeString(bs);                                                \
    }                                                                     \
  }

#define INIT_MRATTR_WITH_LONG(name)                                       \
  if (m_Embed.m_p##name &&                                                \
      UTF8ToLong(m_Embed.m_p##name, SRC_NULL_TERMINATED, &l))             \
  {                                                                       \
    mMidRadio.put_##name(l);                                              \
  }

void nsPluginInstance::InitializeMidRadioAttributes()
{
  if (m_bEmbedded)
  {
    CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);
    BSTR bs;
    long l;
    bool b;

    if (m_Embed.m_pAutoStart &&
        GetBoolFromUTF8(m_Embed.m_pAutoStart, &b))
    {
      mMidRadio.put_AutoStart(b ? VARIANT_TRUE : VARIANT_FALSE);
    }

    if (m_Embed.m_pRepeat)
    {
      const bool bTrue = (_stricmp(m_Embed.m_pRepeat, "true") == 0);
      const bool bFalse = (_stricmp(m_Embed.m_pRepeat, "false") == 0);
      VARIANT v;

      ::VariantInit(&v);
      if (bTrue || bFalse)
      {
        v.vt = VT_BOOL;
        v.boolVal = (bTrue ? VARIANT_TRUE : VARIANT_FALSE);
        mMidRadio.put_Repeat(v);
      }
      else
      {
        l = 0;
        if (UTF8ToLong(m_Embed.m_pRepeat, SRC_NULL_TERMINATED, &l))
        {
          v.vt = VT_I4;
          v.lVal = l;
          mMidRadio.put_Repeat(v);
        }
      }
    }

    INIT_MRATTR_WITH_BSTR(Text)
    INIT_MRATTR_WITH_LONG(Volume)
    INIT_MRATTR_WITH_LONG(TimeFormat)
    INIT_MRATTR_WITH_BSTR(CurrentPosition)
    INIT_MRATTR_WITH_BSTR(StopPosition)
    INIT_MRATTR_WITH_BSTR(Duration)
    INIT_MRATTR_WITH_LONG(Transpose)
    INIT_MRATTR_WITH_LONG(Tempo)
  }
}

void DrawPluginErrMsg(HWND hWnd, HDC hdc, LPPAINTSTRUCT lpps, const TCHAR *lpszErr)
{
  static HBRUSH hBrushWhite = GetStockBrush(WHITE_BRUSH);
  RECT rc;

  HBRUSH hBrushOld = SelectBrush(hdc, hBrushWhite);
  ::PatBlt(hdc,
    lpps->rcPaint.left,
    lpps->rcPaint.top,
    lpps->rcPaint.right - lpps->rcPaint.left,
    lpps->rcPaint.bottom - lpps->rcPaint.top, PATCOPY);
  SelectBrush(hdc, hBrushOld);

  ::GetClientRect(hWnd, &rc);
  ::SetBkColor(hdc, RGB(0xFF, 0xFF, 0xFF));
  ::SetTextColor(hdc, RGB(0xFF, 0, 0));
  ::DrawText(hdc, lpszErr, -1, &rc, DT_WORDBREAK | DT_CENTER);
}

char* NPN_StrDuplicate(const char* s)
{
  return strcpy((char*)NPN_MemAlloc(strlen(s) + 1), s);
}

BSTR AllocSysString(const char *pchData, int cbMultiByte, UINT CodePage)
{
  if (SRC_NULL_TERMINATED == cbMultiByte)
  {
    cbMultiByte = ::lstrlenA(pchData);
  }

  int nLen = ::MultiByteToWideChar(CodePage, 0, pchData, cbMultiByte, NULL, NULL);
  BSTR bstr = ::SysAllocStringLen(NULL, nLen);

  if (bstr != NULL)
  {
    ::MultiByteToWideChar(CodePage, 0, pchData, cbMultiByte, bstr, nLen);
  }
  return bstr;
}

BSTR AllocSysString(const wchar_t *pchData, int cchWideChar)
{
  if (SRC_NULL_TERMINATED == cchWideChar)
  {
    cchWideChar = ::lstrlenW(pchData);
  }

  BSTR bstr = ::SysAllocStringLen(NULL, cchWideChar);

  if (bstr != NULL)
  {
    ::lstrcpyW(bstr, pchData);
  }
  return bstr;
}

char* NPN_AllocString(LPCWSTR pchData, int cchWideChar, UINT CodePage)
{
  if (SRC_NULL_TERMINATED == cchWideChar)
  {
    cchWideChar = ::lstrlenW(pchData);
  }

  int nLen = ::WideCharToMultiByte(CodePage, 0, pchData, cchWideChar, NULL, 0, NULL, NULL);
  char* str = (char*)NPN_MemAlloc(nLen + 1);

  if (str != NULL)
  {
    ::WideCharToMultiByte(CodePage, 0, pchData, cchWideChar, str, nLen, NULL, NULL);
    str[nLen] = '\0';
  }
  return str;
}

LPWSTR AllocWString(LPCSTR pchData, int cbMultiByte, UINT CodePage)
{
  if (SRC_NULL_TERMINATED == cbMultiByte)
  {
    cbMultiByte = ::lstrlenA(pchData);
  }

  int nLen = ::MultiByteToWideChar(CodePage, 0, pchData, cbMultiByte, NULL, NULL);
  wchar_t* wstr = new wchar_t[nLen + 1];

  if (wstr != NULL)
  {
    ::MultiByteToWideChar(CodePage, 0, pchData, cbMultiByte, wstr, nLen);
    wstr[nLen] = L'\0';
  }
  return wstr;
}

BSTR AllocSysStringFromUTF8(LPCSTR lpszUTF8, int cbMultiByte)
{
  BSTR bstr = NULL;
  LPWSTR lpwz = AllocWString(lpszUTF8, cbMultiByte, CP_UTF8);

  if (lpwz)
  {
    bstr = AllocSysString(lpwz, SRC_NULL_TERMINATED);
    delete [] lpwz;
  }
  return bstr;
}

bool UTF8ToLong(const char *lpszUTF8, int cbMultiByte, LONG *plOut)
{
  bool bResult = false;
  LPWSTR lpwz = AllocWString(lpszUTF8, cbMultiByte, CP_UTF8);

  if (lpwz)
  {
    bResult = (::VarI4FromStr(lpwz, 1033, 0, plOut) == S_OK);
    delete [] lpwz;
  }
  return bResult;
}

bool UTF8ToDouble(const char *lpszUTF8, int cbMultiByte, double *pdOut)
{
  bool bResult = false;
  LPWSTR lpwz = AllocWString(lpszUTF8, cbMultiByte, CP_UTF8);

  if (lpwz)
  {
    bResult = (::VarR8FromStr(lpwz, 1033, 0, pdOut) == S_OK);
    delete [] lpwz;
  }
  return bResult;
}

LRESULT CALLBACK
nsPluginInstance::PluginWinProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  nsPluginInstance *plugin = (nsPluginInstance *)::GetProp(hWnd, PLUGIN_INSTANCE_PTR);
  if (!plugin)
  {
    return TRUE;
  }

  switch (msg)
  {
  case WM_PAINT:
    if (!igd.AtlInitialized() ||
        NULL == plugin->m_pUnkControl)
    {
      PAINTSTRUCT ps;
      HDC hdc = ::BeginPaint(hWnd, &ps);

      if (!igd.AtlInitialized())
      {
        DrawPluginErrMsg(hWnd, hdc, &ps, _T("ATL hasn't been initialized!"));
      }
      else if (NULL == plugin->m_pUnkControl)
      {
        DrawPluginErrMsg(hWnd, hdc, &ps, _T("MidRadio hasn't been installed!"));
      }
      ::EndPaint(hWnd, &ps);
    }
    break;
  }
  return ::CallWindowProc(plugin->m_orgPluginWndProc, hWnd, msg, wParam, lParam);
}


// MidRadio events

bool nsPluginInstance::DispatchEvent(const NPUTF8 *lpszEventName)
{
  bool bResult = false;

  if (sWindowObj)
  {
    NPVariant vrDocment;

    if (NPN_GetProperty(m_pInstance, sWindowObj,
                        NPN_GetStringIdentifier("document"), &vrDocment))
    {
      NPVariant vrEvent;
      NPVariant vrType;

      STRINGZ_TO_NPVARIANT("HTMLEvents", vrType);
      if (NPN_Invoke(m_pInstance, NPVARIANT_TO_OBJECT(vrDocment),
                     NPN_GetStringIdentifier("createEvent"),
                     &vrType, 1, &vrEvent))
      {
        NPVariant vrResult;

        NPVariant args[3];
        STRINGZ_TO_NPVARIANT(lpszEventName, args[0]);
        BOOLEAN_TO_NPVARIANT(true, args[1]);     
        BOOLEAN_TO_NPVARIANT(true, args[2]);
        if (NPN_Invoke(m_pInstance, NPVARIANT_TO_OBJECT(vrEvent),
                       NPN_GetStringIdentifier("initEvent"),
                       args, 3, &vrResult))
        {
          NPN_ReleaseVariantValue(&vrResult);

          NPObject* plObj = NULL;
          if (NPN_GetValue(m_pInstance, NPNVPluginElementNPObject, &plObj) == NPERR_NO_ERROR)
          {
            if (NPN_Invoke(m_pInstance, plObj,
                           NPN_GetStringIdentifier("dispatchEvent"),
                           &vrEvent, 1, &vrResult))
            {
              bResult = true;
              NPN_ReleaseVariantValue(&vrResult);
            }

            NPN_ReleaseObject(plObj);
          }
        }

        NPN_ReleaseVariantValue(&vrEvent);
      }

      NPN_ReleaseVariantValue(&vrDocment);
    }
  }
  return bResult;
}


#define DEFINE_PLUGININSTANCE_EVENT(name)                           \
  void nsPluginInstance::##name(void)                               \
  {                                                                 \
    _RPTF0(_CRT_WARN, "nsPluginInstance::"#name"\n");               \
                                                                    \
    DispatchEvent(#name);                                           \
  }

DEFINE_PLUGININSTANCE_EVENT(OnEnd)
DEFINE_PLUGININSTANCE_EVENT(OnEndClip)
DEFINE_PLUGININSTANCE_EVENT(OnStart)
DEFINE_PLUGININSTANCE_EVENT(OnStartClip)
DEFINE_PLUGININSTANCE_EVENT(OnPause)
DEFINE_PLUGININSTANCE_EVENT(OnNext)
DEFINE_PLUGININSTANCE_EVENT(OnPrev)
DEFINE_PLUGININSTANCE_EVENT(OnRestart)
DEFINE_PLUGININSTANCE_EVENT(OnClickPlay)
DEFINE_PLUGININSTANCE_EVENT(OnClickStop)
DEFINE_PLUGININSTANCE_EVENT(OnClickPause)
DEFINE_PLUGININSTANCE_EVENT(OnClickNext)
DEFINE_PLUGININSTANCE_EVENT(OnClickPrev)


#ifdef NPMRHLP_SCRIPTABLE
// ==============================
// ! Scriptability related code !
// ==============================

static bool
NPVariantToBool(const NPVariant *variant)
{
  switch (variant->type)
  {
  case NPVariantType_Void:
  case NPVariantType_Null:
    return false;

  case NPVariantType_Bool:
    return NPVARIANT_TO_BOOLEAN(*variant);

  case NPVariantType_Int32:
    return NPVARIANT_TO_INT32(*variant) != 0;

  case NPVariantType_Double:
    return NPVARIANT_TO_DOUBLE(*variant) != 0.0;

  case NPVariantType_String:
    return NPVARIANT_TO_STRING(*variant).UTF8Length != 0;

  default:
    return true;
  }
}

static bool
NPVariantToInt32(const NPVariant *variant, int32_t *result)
{
  switch (variant->type)
  {
  case NPVariantType_Void:
  case NPVariantType_Null:
    *result = 0;
    return true;

  case NPVariantType_Bool:
    *result = (NPVARIANT_TO_BOOLEAN(*variant) ? 1 : 0);
    return true;

  case NPVariantType_Int32:
    *result = NPVARIANT_TO_INT32(*variant);
    return true;

  case NPVariantType_Double:
    *result = (int32_t)NPVARIANT_TO_DOUBLE(*variant);
    return true;

  case NPVariantType_String:
    if (NPVARIANT_TO_STRING(*variant).UTF8Length == 0)
    {
      *result = 0;
      return true;
    }
    else
    {
      LONG l;

      if (UTF8ToLong(NPVARIANT_TO_STRING(*variant).UTF8Characters,
                     NPVARIANT_TO_STRING(*variant).UTF8Length,
                     &l))
      {
        *result = l;
        return true;
      }
    }
    break;
  }
  return false;
}


NPError	nsPluginInstance::GetValue(NPPVariable aVariable, void *aValue)
{
  NPError rv = NPERR_NO_ERROR;

  switch (aVariable) {
    case NPPVpluginNameString:
      *((char **)aValue) = "MidRadio Helper";
      break;

    case NPPVpluginDescriptionString:
      *((char **)aValue) = "MidRadio Helper plug-in";
      break;

    case NPPVpluginScriptableNPObject:
      *(NPObject **)aValue = GetScriptableObject();
      break;

    default:
      rv = NPERR_GENERIC_ERROR;
      break;
  }
  return rv;
}

NPObject* nsPluginInstance::GetScriptableObject()
{
  if (!m_pScriptableObject) {
    m_pScriptableObject = NPN_CreateObject(m_pInstance,
      GET_NPOBJECT_CLASS(ScriptablePluginObject));
  }

  if (m_pScriptableObject) {
    NPN_RetainObject(m_pScriptableObject);
  }

  return m_pScriptableObject;
}


// ==============================
// ! Scriptability related code !
// ==============================

#define STR_EQUAL(name) (_stricmp(str, #name) == 0)

bool nsPluginInstance::HasMethod(NPIdentifier name)
{
  NPUTF8 *str = NPN_UTF8FromIdentifier(name);

  bool bResult = (STR_EQUAL(Play) ||
                  STR_EQUAL(Stop) ||
                  STR_EQUAL(Pause) ||
                  STR_EQUAL(Next) ||
                  STR_EQUAL(Prev));

  NPN_MemFree(str);
  return bResult;
}

bool nsPluginInstance::HasProperty(NPIdentifier name)
{
  NPUTF8 *str = NPN_UTF8FromIdentifier(name);

  bool bResult = (STR_EQUAL(Src) ||
                  STR_EQUAL(AutoStart) ||
                  STR_EQUAL(Repeat) ||
                  STR_EQUAL(Text) ||
                  STR_EQUAL(Volume) ||
                  STR_EQUAL(Tempo) ||
                  STR_EQUAL(TimeFormat) ||
                  STR_EQUAL(CurrentPosition) ||
                  STR_EQUAL(StopPosition) ||
                  STR_EQUAL(Duration) ||
                  STR_EQUAL(Transpose) ||
                  STR_EQUAL(Version));

  NPN_MemFree(str);
  return bResult;
}

#define SET_MR_PROPERTY(name)     \
  if (_stricmp(str, #name) == 0)  \
  {                               \
    bResult = Set##name(value);   \
  }

bool nsPluginInstance::SetProperty(NPIdentifier name, const NPVariant *value)
{
  NPUTF8 *str = NPN_UTF8FromIdentifier(name);
  bool bResult = false;

  SET_MR_PROPERTY(Src)
  else SET_MR_PROPERTY(AutoStart)
  else SET_MR_PROPERTY(Repeat)
  else SET_MR_PROPERTY(Text) 
  else SET_MR_PROPERTY(Volume)
  else SET_MR_PROPERTY(Tempo)
  else SET_MR_PROPERTY(TimeFormat)
  else SET_MR_PROPERTY(CurrentPosition)
  else SET_MR_PROPERTY(StopPosition)
  else SET_MR_PROPERTY(Duration)
  else SET_MR_PROPERTY(Transpose)
  else SET_MR_PROPERTY(Version)

  NPN_MemFree(str);
  return bResult;
}

#define GET_MR_PROPERTY(name)     \
  if (_stricmp(str, #name) == 0)  \
  {                               \
    bResult = Get##name(result);  \
  }

bool nsPluginInstance::GetProperty(NPIdentifier name, NPVariant *result)
{
  NPUTF8 *str = NPN_UTF8FromIdentifier(name);
  bool bResult = false;

  GET_MR_PROPERTY(Src)
  else GET_MR_PROPERTY(AutoStart)
  else GET_MR_PROPERTY(Repeat)
  else GET_MR_PROPERTY(Text) 
  else GET_MR_PROPERTY(Volume)
  else GET_MR_PROPERTY(Tempo)
  else GET_MR_PROPERTY(TimeFormat)
  else GET_MR_PROPERTY(CurrentPosition)
  else GET_MR_PROPERTY(StopPosition)
  else GET_MR_PROPERTY(Duration)
  else GET_MR_PROPERTY(Transpose)
  else GET_MR_PROPERTY(Version)

  NPN_MemFree(str);
  return bResult;
}

#define INVOKE_MR_METHOD(name)    \
  if (_stricmp(str, #name) == 0)  \
  {                               \
    ##name();                     \
    bResult = true;               \
  }

bool nsPluginInstance::Invoke(NPIdentifier name, const NPVariant *args,
                              uint32_t argCount, NPVariant *result)
{
  NPUTF8 *str = NPN_UTF8FromIdentifier(name);
  bool bResult = false;

  INVOKE_MR_METHOD(Play)
  else INVOKE_MR_METHOD(Stop)
  else INVOKE_MR_METHOD(Pause)
  else INVOKE_MR_METHOD(Next)
  else INVOKE_MR_METHOD(Prev)

  NPN_MemFree(str);
  return bResult;
}

bool nsPluginInstance::InvokeDefault(const NPVariant *args, uint32_t argCount,
                                     NPVariant *result)
{
  _RPTF0(_CRT_WARN, "ScriptablePluginObject default method called!\n");
  return true;
}



// ==============================
// ! Scriptability related code !
// ==============================

#define DEFINE_PLUGININSTANCE_METHOD(name)                              \
  void nsPluginInstance::##name(void)                                   \
  {                                                                     \
    _RPTF0(_CRT_WARN, "nsPluginInstance::"#name"\n");                   \
                                                                        \
    CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);       \
    mMidRadio.raw_##name();                                             \
  }


#define DEFINE_PLUGININSTANCE_PROPERTY_GET_STRING(name)                 \
  bool nsPluginInstance::Get##name(NPVariant *result)                   \
  {                                                                     \
    bool bResult = false;                                               \
    _RPTF0(_CRT_WARN, "nsPluginInstance::Get"#name"\n");                \
                                                                        \
    if (result)                                                         \
    {                                                                   \
      CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);     \
      BSTR bs = NULL;                                                   \
                                                                        \
      if (SUCCEEDED(mMidRadio.get_##name(&bs)) && bs)                   \
      {                                                                 \
        char *lpsz = NPN_AllocString(bs, SRC_NULL_TERMINATED, CP_UTF8); \
        if (lpsz)                                                       \
        {                                                               \
          STRINGN_TO_NPVARIANT(lpsz, strlen(lpsz), *result);            \
          bResult = true;                                               \
        }                                                               \
        ::SysFreeString(bs);                                            \
      }                                                                 \
    }                                                                   \
    return bResult;                                                     \
  }


#define DEFINE_PLUGININSTANCE_PROPERTY_SET_STRING(name)                 \
  bool nsPluginInstance::Set##name(const NPVariant *value)              \
  {                                                                     \
    _RPTF0(_CRT_WARN, "nsPluginInstance::Set"#name"\n");                \
    bool bResult = false;                                               \
                                                                        \
    if (value && NPVARIANT_IS_STRING(*value))                           \
    {                                                                   \
      CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);     \
                                                                        \
      BSTR bs = AllocSysStringFromUTF8(NPVARIANT_TO_STRING(*value).UTF8Characters,  \
                                       NPVARIANT_TO_STRING(*value).UTF8Length);     \
      if (bs)                                                           \
      {                                                                 \
        bResult = SUCCEEDED(mMidRadio.put_##name(bs));                  \
        ::SysFreeString(bs);                                            \
      }                                                                 \
    }                                                                   \
    return bResult;                                                     \
  }


#define DEFINE_PLUGININSTANCE_PROPERTY_GET_INT32(name)                  \
  bool nsPluginInstance::Get##name(NPVariant *result)                   \
  {                                                                     \
    _RPTF0(_CRT_WARN, "nsPluginInstance::Get"#name"\n");                \
    bool bResult = false;                                               \
                                                                        \
    if (result)                                                         \
    {                                                                   \
      CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);     \
      long lVal;                                                        \
                                                                        \
      if (SUCCEEDED(mMidRadio.get_##name(&lVal)))                       \
      {                                                                 \
        INT32_TO_NPVARIANT(lVal, *result);                              \
        bResult = true;                                                 \
      }                                                                 \
    }                                                                   \
    return bResult;                                                     \
  }


#define DEFINE_PLUGININSTANCE_PROPERTY_SET_INT32(name)                  \
  bool nsPluginInstance::Set##name(const NPVariant *value)              \
  {                                                                     \
    _RPTF0(_CRT_WARN, "nsPluginInstance::Set"#name"\n");                \
    bool bResult = false;                                               \
                                                                        \
    if (value)                                                          \
    {                                                                   \
      CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);     \
      int32_t l;                                                        \
                                                                        \
      if (NPVariantToInt32(value, &l))                                  \
      {                                                                 \
        bResult = SUCCEEDED(mMidRadio.put_##name(l));                   \
      }                                                                 \
    }                                                                   \
    return bResult;                                                     \
  }


DEFINE_PLUGININSTANCE_METHOD(Play)
DEFINE_PLUGININSTANCE_METHOD(Stop)
DEFINE_PLUGININSTANCE_METHOD(Pause)
DEFINE_PLUGININSTANCE_METHOD(Next)
DEFINE_PLUGININSTANCE_METHOD(Prev)


DEFINE_PLUGININSTANCE_PROPERTY_GET_STRING(Src)
DEFINE_PLUGININSTANCE_PROPERTY_SET_STRING(Src)

bool nsPluginInstance::GetAutoStart(NPVariant *result)
{
  _RPTF0(_CRT_WARN, "nsPluginInstance::GetAutoStart\n");
  bool bResult = false;

  CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);
  VARIANT_BOOL vb;

  if (SUCCEEDED(mMidRadio.get_AutoStart(&vb)))
  {
    BOOLEAN_TO_NPVARIANT(VARIANT_TRUE == vb, *result);
    bResult = true;
  }
  return bResult;
}

bool nsPluginInstance::SetAutoStart(const NPVariant *value)
{
  _RPTF0(_CRT_WARN, "nsPluginInstance::SetAutoStart\n");
  bool bResult = false;

  if (value)
  {
    CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);
    bResult = SUCCEEDED(mMidRadio.put_AutoStart(NPVariantToBool(value) ? VARIANT_TRUE : VARIANT_FALSE));
  }
  return bResult;
}

bool nsPluginInstance::GetRepeat(NPVariant *result)
{
  _RPTF0(_CRT_WARN, "nsPluginInstance::GetRepeat\n");
  bool bResult = false;

  if (result)
  {
    CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);
    VARIANT vr;

    ::VariantInit(&vr);
    if (SUCCEEDED(mMidRadio.get_Repeat(&vr)))
    {
      if (VT_BOOL == vr.vt)
      {
        BOOLEAN_TO_NPVARIANT(vr.boolVal == VARIANT_TRUE, *result);
        bResult = true;
      }
      else if (VT_I4 == vr.vt)
      {
        INT32_TO_NPVARIANT(vr.lVal, *result);
        bResult = true;
      }
    }
  }
  return bResult;
}

bool nsPluginInstance::SetRepeat(const NPVariant *value)
{
  _RPTF0(_CRT_WARN, "nsPluginInstance::SetRepeat\n");
  bool bResult = false;

  if (value)
  {
    CMidRadio mMidRadio(igd.GetCritSec(), m_pUnkControl, m_hWnd);
    VARIANT vr;

    ::VariantInit(&vr);
    switch (value->type)
    {
    case NPVariantType_Void:
    case NPVariantType_Null:
      vr.vt = VT_BOOL;
      vr.boolVal = VARIANT_FALSE;
      bResult = SUCCEEDED(mMidRadio.put_Repeat(vr));
      break;

    case NPVariantType_Bool:
      vr.vt = VT_BOOL;
      vr.boolVal = (NPVARIANT_TO_BOOLEAN(*value) ? VARIANT_TRUE : VARIANT_FALSE);
      bResult = SUCCEEDED(mMidRadio.put_Repeat(vr));
      break;

    case NPVariantType_Int32:
      vr.vt = VT_I4;
      vr.lVal = NPVARIANT_TO_INT32(*value);
      bResult = SUCCEEDED(mMidRadio.put_Repeat(vr));
      break;

    case NPVariantType_Double:
      {
        LONG lOut;

        if (::VarI4FromR8(NPVARIANT_TO_DOUBLE(*value), &lOut) == S_OK)
        {
          vr.vt = VT_I4;
          vr.lVal = lOut;
          bResult = SUCCEEDED(mMidRadio.put_Repeat(vr));
        }
      }
      break;

    case NPVariantType_String:
      vr.vt = VT_BOOL;
      vr.boolVal = ((NPVARIANT_TO_STRING(*value).UTF8Length != 0) ? VARIANT_TRUE : VARIANT_FALSE);
      bResult = SUCCEEDED(mMidRadio.put_Repeat(vr));
      break;

    default:
      vr.vt = VT_BOOL;
      vr.boolVal = VARIANT_TRUE;
      bResult = SUCCEEDED(mMidRadio.put_Repeat(vr));
      break;
    }
  }
  return bResult;
}

DEFINE_PLUGININSTANCE_PROPERTY_GET_STRING(Text)
DEFINE_PLUGININSTANCE_PROPERTY_SET_STRING(Text)

DEFINE_PLUGININSTANCE_PROPERTY_GET_INT32(Volume)
DEFINE_PLUGININSTANCE_PROPERTY_SET_INT32(Volume)

DEFINE_PLUGININSTANCE_PROPERTY_GET_INT32(Tempo)
DEFINE_PLUGININSTANCE_PROPERTY_SET_INT32(Tempo)

DEFINE_PLUGININSTANCE_PROPERTY_GET_INT32(TimeFormat)
DEFINE_PLUGININSTANCE_PROPERTY_SET_INT32(TimeFormat)

DEFINE_PLUGININSTANCE_PROPERTY_GET_STRING(CurrentPosition)
DEFINE_PLUGININSTANCE_PROPERTY_SET_STRING(CurrentPosition)

DEFINE_PLUGININSTANCE_PROPERTY_GET_STRING(StopPosition)
DEFINE_PLUGININSTANCE_PROPERTY_SET_STRING(StopPosition)

DEFINE_PLUGININSTANCE_PROPERTY_GET_STRING(Duration)
DEFINE_PLUGININSTANCE_PROPERTY_SET_STRING(Duration)

DEFINE_PLUGININSTANCE_PROPERTY_GET_INT32(Transpose)
DEFINE_PLUGININSTANCE_PROPERTY_SET_INT32(Transpose)

DEFINE_PLUGININSTANCE_PROPERTY_GET_STRING(Version)

bool nsPluginInstance::SetVersion(const NPVariant *value)
{
  return true;
}

#endif
