/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is 
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef __PLUGIN_H__
#define __PLUGIN_H__

#include "nsCOMPtr.h"
#include "nsIURI.h"

#include "pluginbase.h"
#include "npruntime.h"

#ifdef NPMRHLP_SCRIPTABLE
#include "nsIVariant.h"
#endif
#include "MidRadioEventListener.h"

class nsPluginInstance : public nsPluginInstanceBase
{
public:
  nsPluginInstance(NPP aInstance);
  ~nsPluginInstance();

  NPBool init(NPWindow* aWindow);
  void shut();
  NPBool isInitialized();
  virtual NPError SetWindow(NPWindow* pNPWindow);
  NPError NewStream(NPMIMEType type, NPStream* stream, NPBool seekable, uint16* stype);
  int32_t WriteReady(NPStream *stream);

  void SetPluginCreateData(nsPluginCreateData * lppcd);
  void InitializeMidRadioAttributes();
  BSTR AllocMidRadioCompatibleSrc(LPCWSTR lpwzSrc);
  static LRESULT CALLBACK PluginWinProc(HWND, UINT, WPARAM, LPARAM);

  // MidRadio events
  void OnEnd();
  void OnEndClip();
  void OnStart();
  void OnStartClip();
  void OnPause();
  void OnNext();
  void OnPrev();
  void OnRestart();
  void OnClickPlay();
  void OnClickStop();
  void OnClickPause();
  void OnClickNext();
  void OnClickPrev();

#ifdef NPMRHLP_SCRIPTABLE
  virtual bool HasMethod(NPIdentifier name);
  virtual bool HasProperty(NPIdentifier name);
  virtual bool SetProperty(NPIdentifier name, const NPVariant *value);
  virtual bool GetProperty(NPIdentifier name, NPVariant *result);
  virtual bool Invoke(NPIdentifier name, const NPVariant *args, uint32_t argCount, NPVariant *result);
  virtual bool InvokeDefault(const NPVariant *args, uint32_t argCount, NPVariant *result);

  // we need to provide implementation of this method as it will be
  // used by Mozilla to retrive the scriptable peer
  NPError	GetValue(NPPVariable variable, void *value);

  // locals
  void Play(void);
  void Stop(void);
  void Pause(void);
  void Next(void);
  void Prev(void);
  bool GetSrc(NPVariant *result);
  bool SetSrc(const NPVariant *value);
  bool GetAutoStart(NPVariant *result);
  bool SetAutoStart(const NPVariant *value);
  bool GetRepeat(NPVariant *result);
  bool SetRepeat(const NPVariant *value);
  bool GetText(NPVariant *result);
  bool SetText(const NPVariant *value);
  bool GetVolume(NPVariant *result);
  bool SetVolume(const NPVariant *value);
  bool GetTempo(NPVariant *result);
  bool SetTempo(const NPVariant *value);
  bool GetTimeFormat(NPVariant *result);
  bool SetTimeFormat(const NPVariant *value);
  bool GetCurrentPosition(NPVariant *result);
  bool SetCurrentPosition(const NPVariant *value);
  bool GetStopPosition(NPVariant *result);
  bool SetStopPosition(const NPVariant *value);
  bool GetDuration(NPVariant *result);
  bool SetDuration(const NPVariant *value);
  bool GetTranspose(NPVariant *result);
  bool SetTranspose(const NPVariant *value);
  bool GetVersion(NPVariant *result);
  bool SetVersion(const NPVariant *value);
#endif

private:
#ifdef NPMRHLP_SCRIPTABLE
  NPObject *GetScriptableObject();
  bool DispatchEvent(const NPUTF8 *lpszEventName);

  NPObject *m_pScriptableObject;

  // Methods
  NPIdentifier m_id_Play;
  NPIdentifier m_id_Stop;
  NPIdentifier m_id_Pause;
  NPIdentifier m_id_Next;
  NPIdentifier m_id_Prev;

  // Properties
  NPIdentifier m_id_Src;
  NPIdentifier m_id_AutoStart;
  NPIdentifier m_id_Repeat;
  NPIdentifier m_id_Text;
  NPIdentifier m_id_Volume;
  NPIdentifier m_id_Tempo;
  NPIdentifier m_id_TimeFormat;
  NPIdentifier m_id_CurrentPosition;
  NPIdentifier m_id_StopPosition;
  NPIdentifier m_id_Duration;
  NPIdentifier m_id_Transpose;
  NPIdentifier m_id_Version;
#endif

  NPP m_pInstance;
  NPObject* sWindowObj;
  NPBool m_bInitialized;
  HWND m_hWnd;
  WNDPROC m_orgPluginWndProc;
  bool m_bComInit;

  HWND m_hWndContainer;
  IUnknown* m_pUnkControl;

  CMidRadioEventListener * m_pMidRadioEventListener;

  bool m_bEmbedded;

  // embedded data
  struct EmbeddedData
  {
    char* m_pAutoStart;
    char* m_pRepeat;
    char* m_pText;
    char* m_pVolume;
    char* m_pTempo;
    char* m_pTimeFormat;
    char* m_pCurrentPosition;
    char* m_pStopPosition;
    char* m_pDuration;
    char* m_pTranspose;
  } m_Embed;
};

#endif // __PLUGIN_H__
